#include <string.h>
#include "FileOutput.h"

/*
 * An object which writes simultaneously to multiple outputs. Applies the singleton
 * design pattern.
 * Output may consist of any of int,char,char*,float,double and Printable. Printable
 * is an interface which defines the pure virtual function 'toString()' which returns
 * the string representation of an object (similar to the convention found in
 * the java programming language).
 *
 * Author: Ziv Yaniv (zivy@cs.huji.ac.il).
 */

FileOutput * FileOutput::_instance = NULL;

/*
 * Create the FileOutput object and add stdout and stderr to the maps.
 */
FileOutput::FileOutput() : STDOUT(1), STDERR(2)
{
	FileData *fd;  

  fd = new FileData("stdout",STDOUT);
  fd->_f = stdout;
  _filesByName.insert(
		       map<char *, FileData *,StrComparator>::value_type("stdout",fd));
  _filesById.insert(
		       map<unsigned long, FileData *>::value_type(STDOUT,fd));

  fd = new FileData("stderr",STDERR);
  fd->_f = stderr;
   _filesByName.insert(
		       map<char *, FileData *,StrComparator>::value_type("stderr",fd));
  _filesById.insert(
		       map<unsigned long, FileData *>::value_type(STDERR,fd));

  _freeFileIds = 0xffffffff^STDOUT^STDERR;
}
/*****************************************************************************/
/*
 * Get a pointer to the only instance of a FileOutput object.
 */
FileOutput *FileOutput::getInstance()
{
	if(_instance == NULL) 
    _instance = new FileOutput();
	return(_instance);
}
/*****************************************************************************/
/*
 * Add a file to the FileOutput object. Every file has a unique id 
 * (unsigned long).
 * Returns: id > 0 if successful.
 *          0 upon failure.
 */
unsigned long FileOutput::addFile(char *name,const char *mode)
{
	FileData *fd;
  FILE *f;
  unsigned long id;
  
        //file with this name already exists
  if(_filesByName.find(name) != _filesByName.end())
    return(0);
  
  if((f = fopen(name,mode)) == NULL) 
    return(0);

          //search for the first free file id 
  id = 1;  
  while(!(id & _freeFileIds))
    id = id<<1;
  _freeFileIds^= id; 

  fd = new FileData(name,id);
  fd->_f = f;
  _filesByName.insert(
		       map<char *, FileData *,StrComparator>::value_type(name,fd));
  _filesById.insert(
		       map<unsigned long, FileData *>::value_type(id,fd));
  return(id);
}
/*****************************************************************************/
/*
 * Given a file I.D. returns the files name. This function allocates memory
 * for the string. The user must free this memory using 'free'.
 */
char *FileOutput::fileName(unsigned long fileId)
{
	map<unsigned long, FileData *>::iterator idIter;

	idIter = _filesById.find(fileId);
	
	if(idIter == _filesById.end())
		return(NULL);
	return(_strdup((*idIter).second->_name));
}
/*****************************************************************************/
/*
 * This method closes all files and frees the memory associated with them.
 */
void FileOutput::closeFiles(unsigned long fileIds)
{
  unsigned long id = 1;
  map<unsigned long, FileData *>::iterator idIter;
  map<char *, FileData *,StrComparator>::iterator nameIter;
  FileData *fd; 

	     //can't close stderr or stdout
	if(fileIds & STDERR)
		fileIds^= STDERR;
	if(fileIds & STDOUT)
		fileIds^= STDOUT;

  do {         //close files which are specified by 'fileIds'
               //and aren't stdout or stderr
    if(id & fileIds) {
      idIter = _filesById.find(id);
      nameIter = _filesByName.find((*idIter).second->_name);
      fd = (*idIter).second;
      _filesById.erase(idIter);
      _filesByName.erase(nameIter);
      fclose(fd->_f);
      delete fd;

			fileIds^= id;
      _freeFileIds|= id;
    }
    id = id<<1;
  } while(fileIds);
}
/*****************************************************************************/
/*
 * Write an int to the specified files.
 */
void FileOutput::print(unsigned long fileIds, int i)
{
  unsigned long id = 1;
  map<unsigned long, FileData *>::iterator idIter;

  do {
    if(id & fileIds) {
      idIter = _filesById.find(id);
      fprintf((*idIter).second->_f,"%d",i);
      fileIds^= id;
    }
    id = id<<1;
  } while(fileIds);
}
/*****************************************************************************/
/*
 * Write an int to the specified files followed by a newline.
 */
void FileOutput::println(unsigned long fileIds, int i)
{
  unsigned long id = 1;
  map<unsigned long, FileData *>::iterator idIter;

  do {
    if(id & fileIds) {
      idIter = _filesById.find(id);
      fprintf((*idIter).second->_f,"%d\n",i);
      fileIds^= id;
    }
    id = id<<1;
  } while(fileIds);
}
/*****************************************************************************/
/*
 * Write a character to the specified files.
 */
void FileOutput::print(unsigned long fileIds, char c)
{
  unsigned long id = 1;
  map<unsigned long, FileData *>::iterator idIter;

  do {
    if(id & fileIds) {
      idIter = _filesById.find(id);
      fprintf((*idIter).second->_f,"%c",c);
      fileIds^= id;
    }
    id = id<<1;
  } while(fileIds);
}
/*****************************************************************************/
/*
 * Write a character to the specified files followed by a newline.
 */
void FileOutput::println(unsigned long fileIds, char c)
{
  unsigned long id = 1;
  map<unsigned long, FileData *>::iterator idIter;

  do {
    if(id & fileIds) {
      idIter = _filesById.find(id);
      fprintf((*idIter).second->_f,"%c\n",c);
      fileIds^= id;
    }
    id = id<<1;
  } while(fileIds);
}
/*****************************************************************************/
/*
 * Write a string to the specified files.
 */
void FileOutput::print(unsigned long fileIds, char *s)
{
  unsigned long id = 1;
  map<unsigned long, FileData *>::iterator idIter;

  do {
    if(id & fileIds) {
      idIter = _filesById.find(id);
      fprintf((*idIter).second->_f,"%s",s);
      fileIds^= id;
    }
    id = id<<1;
  } while(fileIds);
}
/*****************************************************************************/
/*
 * Write a string to the specified files followed by a newline.
 */
void FileOutput::println(unsigned long fileIds, char *s)
{
  unsigned long id = 1;
  map<unsigned long, FileData *>::iterator idIter;

  do {
    if(id & fileIds) {
      idIter = _filesById.find(id);
      fprintf((*idIter).second->_f,"%s\n",s);
      fileIds^= id;
    }
    id = id<<1;
  } while(fileIds);
}
/*****************************************************************************/
/*
 * Write a float to the specified files.
 */
void FileOutput::print(unsigned long fileIds, float f)
{
  unsigned long id = 1;
  map<unsigned long, FileData *>::iterator idIter;

  do {
    if(id & fileIds) {
      idIter = _filesById.find(id);
      fprintf((*idIter).second->_f,"%f",f);
      fileIds^= id;
    }
    id = id<<1;
  } while(fileIds);
}
/*****************************************************************************/
/*
 * Write a float to the specified files followed by a newline.
 */
void FileOutput::println(unsigned long fileIds, float f)
{
  unsigned long id = 1;
  map<unsigned long, FileData *>::iterator idIter;

  do {
    if(id & fileIds) {
      idIter = _filesById.find(id);
      fprintf((*idIter).second->_f,"%f\n",f);
      fileIds^= id;
    }
    id = id<<1;
  } while(fileIds);
}
/*****************************************************************************/
/*
 * Write a double to the specified files.
 */
void FileOutput::print(unsigned long fileIds, double d)
{
  unsigned long id = 1;
  map<unsigned long, FileData *>::iterator idIter;

  do {
    if(id & fileIds) {
      idIter = _filesById.find(id);
      fprintf((*idIter).second->_f,"%lg",d);
      fileIds^= id;
    }
    id = id<<1;
  } while(fileIds);
}
/*****************************************************************************/
/*
 * Write a double to the specified files followed by a newline.
 */
void FileOutput::println(unsigned long fileIds, double d)
{
  unsigned long id = 1;
  map<unsigned long, FileData *>::iterator idIter;

  do {
    if(id & fileIds) {
      idIter = _filesById.find(id);
      fprintf((*idIter).second->_f,"%lg\n",d);
      fileIds^= id;
    }
    id = id<<1;
  } while(fileIds);
}


/*****************************************************************************/
/*
 * Write a Printable object to the specified files.
 */
void FileOutput::print(unsigned long fileIds, Printable &p)
{
  char *c = p.toString();
  print(fileIds,c);
}
/*****************************************************************************/ 
/*
 * Write a Printable object to the specified files followed by a newline.
 */
void FileOutput::println(unsigned long fileIds, Printable &p)
{
  char *c = p.toString();
  println(fileIds,c);
}
